
# NOTE: This string is placed into the top of the catalog JSON file as a JSON string.  JSON strings
# are enclosed in double quotes, so it would quickly get ugly if we used double quotes within this
# description, even though python makes that easy.
catalog_file_description = """
    This JSON file has the following format.  Comments are, of course, not present (since JSON does not support
    comments).  Single quotes here are, of course, double quotes in the rest of the file (since JSON encloses
    strings in double quotes).  Anything inside <angle brackets> is just standing in for the relevant value.  An
    ellipsis ... indicates that the preceding structure can be repeated.  Also note that the metadata entries for
    simulations may not be present if the record on zenodo is closed-access; see catalog_private_metadata.json if
    you have access to those simulations, which will contain the missing information.  That file should be read
    and written automatically by functions in this module, so that the catalog dict returned will contain all
    available information.

    {
        'catalog_file_description': '<this description>',
        'modified': '<YYYY-MM-DDThh:mm:ss.ssssss>',  # UTC time of last-modified record in this file
        'records': {  # Includes *all* records published on Zenodo in the 'sxs' community, not just simulations
            '<doi_url>': {  # This is precisely the value of the 'doi_url' key below
                # More details about this 'representation' object at http://developers.zenodo.org/#depositions
                'conceptdoi': '10.5281/zenodo.<conceptrecid>',  # Permanent DOI for all versions of this record
                'conceptrecid': '<conceptrecid>',  # ~7-digit integer (as string) collectively identifying all versions of this record
                'created': '<YYYY-MM-DDThh:mm:ss.ssssss>',  # UTC time of creation of this record on Zenodo
                'doi': '10.5281/zenodo.<id>',  # Permanent DOI for this record
                'doi_url': 'https://doi.org/10.5281/zenodo.<id>',  # URL for permanent DOI of this record
                'id': <id>,  # ~7-digit integer uniquely identifying this record
                'links': {
                     'badge': 'https://zenodo.org/badge/doi/10.5281/zenodo.<id>.svg',
                     'bucket': 'https://zenodo.org/api/files/<uuid>',  # Base URL for file uploads and downloads
                     'conceptbadge': 'https://zenodo.org/badge/doi/10.5281/zenodo.<conceptrecid>.svg',
                     'conceptdoi': 'https://doi.org/10.5281/zenodo.<conceptrecid>',  # Permanent link to webpage for most-recent version
                     'discard': 'https://zenodo.org/api/deposit/depositions/<id>/actions/discard',  # API action to discard a draft
                     'doi': 'https://doi.org/10.5281/zenodo.<id>',  # Permanent URL for this version
                     'edit': 'https://zenodo.org/api/deposit/depositions/<id>/actions/edit',  # API action to edit this record
                     'files': 'https://zenodo.org/api/deposit/depositions/<id>/files',  # Only present for author
                     'html': 'https://zenodo.org/deposit/<id>',  # Webpage for this version
                     'latest': 'https://zenodo.org/api/records/<id>',  # API endpoint for most-recent version
                     'latest_html': 'https://zenodo.org/record/<id>',  # Webpage for most-recent version
                     'publish': 'https://zenodo.org/api/deposit/depositions/<id>/actions/publish',  # Only present for author
                     'record': 'https://zenodo.org/api/records/<id>',  # Only present for author
                     'record_html': 'https://zenodo.org/record/<id>',  # Webpage for this particular version; only present for author
                     'self': 'https://zenodo.org/api/deposit/depositions/<id>'
                },
                'metadata': {  # Note that this is Zenodo metadata, and is different from the SXS metadata found below
                    'access_right': '<access>',  # Can be 'open', 'closed', 'embargoed', or 'restricted'
                    'communities': [
                        {'identifier': '<community_name>'},  # Names may include 'sxs' and 'zenodo'
                        ...
                    ],
                    'creators': [
                        {
                            'name': '<name>',  # Name of this creator in the format Family name, Given names
                            'affiliation': '<affiliation>',  # (Optional) Affiliation of this creator
                            'orcid': '<orcid>',  # (Optional) ORCID identifier of this creator
                            'gnd': '<gnd>'  # (Optional) GND identifier of this creator
                        },
                        ...
                    ],
                    'description': '<description>',  # Text description of this record
                    'doi': '10.5281/zenodo.<id>',  # Permanent DOI of this record
                    'keywords': [
                        '<keyword>',  # Optional; this array may be empty
                        ...
                    ],
                    'license': '<license_type>',  # Usually 'CC-BY-4.0' for SXS
                    'prereserve_doi': {'doi': '10.5281/zenodo.<id>', 'recid': <id>},
                    'publication_date': '<YYYY-MM-DD>',  # Possibly meaningless date (UTC)
                    'title': '<title>',
                    'upload_type': 'dataset'
                },
                'modified': '<YYYY-MM-DDThh:mm:ss.ssssss>',  # (UTC) Last modification of this record (possibly just Zenodo metadata modified)
                'owner': <user_id>,  # ~5-digit integer identifying the user who owns this record
                'record_id': <id>,  # Same as 'id'
                'state': '<state>',  # Can be 'done', 'inprogress', 'error', 'unsubmitted', possibly others
                'submitted': <submitted>,  # True or false (always true for published records)
                'title': '<title>'  # Same as ['metadata']['title'],
                'files': [  # May not be present if this simulation is closed-access; see catalog_private_metadata.json as noted above
                    # See https://data.black-holes.org/waveforms/documentation.html for
                    # detailed descriptions of the *contents* of the files in each record.
                    {
                        'checksum': '<checksum>',  # MD5 checksum of file on Zenodo
                        'filename': '<filename>',  # Name of file; may contain slashes denoting directories
                        'filesize': <filesize>,  # Number of bytes in the file
                        'id': '<fileid>',  # A standard UUID (hexadecimal with characters in the pattern 8-4-4-4-12)
                        'links': {
                            'download': 'https://zenodo.org/api/files/<bucket>/<filename>',  # The URL to use to download this file
                            'self': 'https://zenodo.org/api/deposit/depositions/<deposition_id>/files/<fileid>'  # Ignore this
                        }
                    },
                    ...  # Other file descriptions in the order in which they were uploaded (not necessarily a meaningful order)
                ]
            },
            ...
        },
        'simulations': {  # Physical data (masses, spins, etc.) for all available SXS simulations
            '<sxs_id>': {  # The SXS ID is a string like SXS:BHNS:0001 or SXS:BBH:1234
                'url': '<URL>',  # The URL of the Zenodo 'conceptdoi' link, which *resolves to* the most-recent version
                #
                # NOTE: All of the following may be absent if this simulation is closed-access, or simply does not have metadata.
                #
                # Variable content describing (mostly) physical parameters of the system.  It's basically a
                # python-compatible version of the information contained in 'metadata.txt' from the
                # highest-resolution run in the most-recent version of this simulation.  That file is meant to
                # be more-or-less as suggested in <https://arxiv.org/abs/0709.0093>.  The conversion to a
                # python-compatible format means that keys like 'simulation-name' have had hyphens replaced by
                # underscores so that they can be used as variable names in python and any other sane language
                # (with apologies to Lisp).  As far as possible, values that are just strings in that file
                # have been converted into the relevant types -- like numbers, integers, and arrays.  Note
                # that some keys like eccentricity are sometimes numbers and sometimes the string '<number'
                # (meaning that the eccentricity is less than the number), which is necessarily a string.
                #
                # Below are just the first few keys that *may* be present.  Note that closed-access
                # simulations will have empty dictionaries here.
                #
                'simulation_name': '<directory_name>',  # This may be distinctly uninformative
                'alternative_names': '<sxs_id>',  # This may be a list of strings
                'initial_data_type': '<type>',  # Something like 'BBH_CFMS'
                'object_types': '<type>',  # Currently 'BHBH', 'BHNS', or 'NSNS'
                'number_of_orbits': <number>,  # This is a float, rather than an integer
                'reference_mass_ratio': <q>,  # Usually greater than 1 (exceptions are due to junk radiation)
                'reference_chi_eff': <chi_eff>,  # Dimensionless effective spin quantity
                'reference_chi1_perp': <chi1_perp>,  # Magnitude of component of chi1 orthogonal to 'reference_orbital_frequency'
                'reference_chi2_perp': <chi2_perp>,  # Magnitude of component of chi2 orthogonal to 'reference_orbital_frequency'
                'reference_mass1': <m2>,
                'reference_mass2': <m1>,
                'reference_dimensionless_spin1': [
                    <chi1_x>,
                    <chi1_y>,
                    <chi1_z>
                ],
                'reference_dimensionless_spin2': [
                    <chi2_x>,
                    <chi2_y>,
                    <chi2_z>
                ],
                'reference_eccentricity': <eccentricity>,  # A float or possibly a string containing '<' and a float
                'reference_orbital_frequency': [
                    <omega_x>,
                    <omega_y>,
                    <omega_z>
                ],
                'reference_time': <time>,
                ...
            },
            ...
        }
    }
"""


def _include_file_data(login, record):
    """Ensure that 'files' field is present in the input record"""
    if 'files' not in record:
        if 'files' in record.get('links', {}):
            url = record['links']['files']
        elif 'id' in record:
            url = login.base_url + 'api/deposit/depositions/{0}/files'.format(record['id'])
        else:
            return record  # We have failed, but we're not too upset
        r = login.session.get(url)
        if r.status_code == 200:  # Otherwise, we just couldn't add this record; oh well.
            r_json = r.json()
            record['files'] = r_json
    return record


def _create_simulations_skeleton(record_list):
    """Create a dictionary of simulations with information for downloading SXS metadata"""
    from .. import sxs_id
    return {
        sxs_id(r.get('title', '')): {
            'url': r['links']['conceptdoi'],
            'metadata_file_info': max([f for f in r.get('files', []) if '/metadata.json' in f['filename']],
                                      default={}, key=lambda f: f['filename'])
        }
        for r in record_list if sxs_id(r.get('title', ''))
    }


def _download_sxs_metadata(login, simulations):
    """Download metadata.json from Zenodo for each simulation, if needed"""
    import traceback
    from tqdm.auto import tqdm
    for sxs_id in tqdm(simulations, dynamic_ncols=True):
        download_url = simulations[sxs_id].get('metadata_file_info', {}).get('links', {}).get('download', '')
        if not download_url:
            continue
        try:
            metadata = login.session.get(download_url).json()
            simulations[sxs_id].pop('metadata_file_info', {})
            simulations[sxs_id].update(metadata)
        except KeyboardInterrupt:
            raise
        except:
            traceback.print_exc()


def create(login=None):
    """Create the catalog from scratch

    This function will take quite some time (more than 15 minutes), because it has
    to download each metadata file individually, which necessarily requires a
    separate Zenodo request for each download.

    """
    from textwrap import dedent
    from . import Login

    # If login is None, this creates a Login object to use
    l = login or Login()

    # Get the list of all SXS records from Zenodo
    records = l.search(q='communities:sxs', all_versions=True, status='published', size=9000)

    # Closed-access records don't have their files included by default; add them
    records = [_include_file_data(l, record) for record in records]

    # Sort the list of records by title
    records = sorted(records, key=lambda r: r.get('title', ''))
    
    # Find the most recent modification time in any record
    modified = max(r.get('modified', '') for r in records)

    # Make an outline of the 'simulations' dict, with info for highest-Lev metadata.json file in
    # place of the actual metadata.
    simulations = _create_simulations_skeleton(records)

    # Loop through the dictionary we just created, and download the metadata.json for each one
    _download_sxs_metadata(l, simulations)

    # Create the catalog itself
    catalog = {
        'catalog_file_description': dedent(catalog_file_description).split('\n')[1:-1],
        'modified': modified,
        'records': {r['doi_url']: r for r in records},
        'simulations': simulations
    }
    return catalog


def split_catalog(catalog):
    """Split catalog four ways: private/public, and complete/simulations

    This function splits the catalog into four separate parts:

      1) The complete catalog for open-access systems
      2) The SXS metadata for open-access systems
      3) The complete catalog for all systems
      4) The SXS metadata for all systems

    The complete catalog contains all Zenodo metadata and SXS metadata, as well as
    a description of the catalog and the most recent modification time.  The
    simulations are useful for serving data used in the waveform index table at
    <https://data.black-holes.org/waveforms/catalog.html>, because it is
    significantly less data.  The private data can be served via the wiki, with the
    public data as a fallback if the user's web browser does not send credentials
    along with the request for the data.

    """
    import copy
    private_catalog = copy.deepcopy(catalog)
    private_simulations = copy.deepcopy(catalog['simulations'])
    public_catalog = {
        'catalog_file_description': catalog['catalog_file_description'],
        'modified': catalog['modified'],
        'records': {
            recid: catalog['records'][recid]
            for recid in catalog['records']
            if catalog['records'][recid].get('metadata', {}).get('access_right', '') == 'open'
        }
    }
    public_conceptdoi_links = [public_catalog['records'][recid].get('links', {}).get('conceptdoi', None)
                               for recid in public_catalog['records']]
    public_catalog['simulations'] = {
        sxs_id: catalog['simulations'][sxs_id]
        for sxs_id in catalog['simulations']
        if catalog['simulations'][sxs_id].get('url', '') in public_conceptdoi_links
    }
    public_simulations = copy.deepcopy(public_catalog['simulations'])
    return private_catalog, private_simulations, public_catalog, public_simulations


def write_split_catalogs(catalog, public_dir='~/.sxs/catalog/', private_dir='~/.sxs/catalog/',
                         public_prefix='', private_prefix=''):
    """Write four versions of the catalog to files

    The input catalog is first split into four parts by `split_catalog` (see that
    function's docstring for details).  These will be written into four separate
    JSON files.  If the public and private directories input as arguments to this
    function are the same, the files will be named private_catalog.json,
    private_simulations.json, public_catalog.json, and public_simulations.json.
    Otherwise, the prefixes will be removed so that the files are named either
    catalog.json or simulations.json when placed in their respective directories.

    Parameters
    ----------
    catalog : dict
        The catalog in the standard format output by functions in this submodule,
        and described in `sxs.zenodo.catalog.catalog_file_description`.
    public_dir : str, optional [defaults to '~/.sxs/catalog/']
    private_dir: str, optional [defaults to '~/.sxs/catalog/']
        Absolute or relative paths to the directory in which the files will be
        written.  If they do not exist, they will be created.  Note that, as
        explained above, the output file names will depend on whether these two
        arguments are the same or different.
    public_prefix: str, optional [defaults to '']
    private_prefix: str, optional [defaults to '']
        Strings to be prepended to the output file names.  If these are empty
        strings but the output directories are the same, these are set to 'public_'
        and 'private_', respectively.

    """
    import os
    from os.path import expanduser, normpath, join
    from datetime import datetime
    import json

    def mkdirs(path):
        # In python >3.2, this could just be os.makedirs(path, exist_ok=True)
        import errno
        import os
        try:
            os.makedirs(path)
        except OSError as exc:
            if exc.errno == errno.EEXIST and os.path.isdir(path):
                pass
            else:
                raise

    public_dir = normpath(expanduser(public_dir))
    private_dir = normpath(expanduser(private_dir))
    mkdirs(public_dir)
    mkdirs(private_dir)

    if public_dir == private_dir:
        if not public_prefix:
            public_prefix = 'public_'
        if not private_prefix:
            private_prefix = 'private_'
    
    private_catalog, private_simulations, public_catalog, public_simulations = split_catalog(catalog)

    public_catalog_path = join(public_dir, public_prefix+'catalog.json')
    public_simulations_path = join(public_dir, public_prefix+'simulations.json')
    private_catalog_path = join(private_dir, private_prefix+'catalog.json')
    private_simulations_path = join(private_dir, private_prefix+'simulations.json')
    with open(public_catalog_path, 'w') as f:
        json.dump(public_catalog, f, indent=2, separators=(',', ': '), ensure_ascii=True)
    with open(public_simulations_path, 'w') as f:
        json.dump(public_simulations, f, indent=None, separators=(',', ':'), ensure_ascii=True)
    with open(private_catalog_path, 'w') as f:
        json.dump(private_catalog, f, indent=2, separators=(',', ': '), ensure_ascii=True)
    with open(private_simulations_path, 'w') as f:
        json.dump(private_simulations, f, indent=None, separators=(',', ':'), ensure_ascii=True)

    public_modified = max(public_catalog['records'][doi].get('modified', '') for doi in public_catalog['records'])
    private_modified = max(private_catalog['records'][doi].get('modified', '') for doi in private_catalog['records'])
    if public_modified:
        public_modified = int(datetime.timestamp(datetime.strptime(public_modified, '%Y-%m-%dT%H:%M:%S.%f')))
        times = (public_modified, public_modified)
        os.utime(public_catalog_path, times)
        os.utime(public_simulations_path, times)
    if private_modified:
        private_modified = int(datetime.timestamp(datetime.strptime(private_modified, '%Y-%m-%dT%H:%M:%S.%f')))
        times = (private_modified, private_modified)
        os.utime(private_catalog_path, times)
        os.utime(private_simulations_path, times)


def update(login=None, path='~/.sxs/catalog/private_catalog.json',
           public_out_dir='~/.sxs/catalog/', private_out_dir='~/.sxs/catalog/',
           nginx_map_file_path='~/.sxs/catalog/sxs_to_zenodo.map'):
    """Update a local copy of the SXS catalog from Zenodo

    Warning: This function works in place, by overwriting the input catalog file,
    as well as the other three versions of it produced by `split_catalog`.

    Parameters
    ----------
    login : sxs.zenodo.Login object, optional [defaults to new Login object]
        This is used to query Zenodo for new data.
    path : str, optional [defaults to '~/.sxs/catalog/private_catalog.json']
        Absolute or relative path to JSON file containing the complete catalog.  If the path does
        not end with '.json', it is assumed to be a directory containing a 'catalog.json' or
        'private_catalog.json' file.
    public_out_dir : str, optional [defaults to '~/.sxs/catalog/']
    private_out_dir : str, optional [defaults to '~/.sxs/catalog/']
        These are passed to the `save_split_catalogs` function; see that function's docstring for
        details.
    nginx_map_file_path : str, optional [defaults to '~/.sxs/catalog/sxs_to_zenodo.map']
        This is passed to the `nginx_map` function, which produces a map file to be used by nginx to
        redirect requests for, e.g., https://data.black-holes.org/SXS:BBH:0001 to the appropriate
        Zenodo concept-doi page.

    """
    from os.path import expanduser, join, dirname, exists
    import json
    from . import Login
    from .. import sxs_id

    # Make sure that the catalog file exists
    path = expanduser(path)
    if not path.endswith('.json'):
        path = join(path, 'private_catalog.json')
        if not exists(path):
            path = join(path, 'catalog.json')
    if not exists(path):
        raise ValueError("Could not find catalog file in '{0}'.".format(dirname(path)))

    # Read the catalog file
    with open(path, 'r') as f:
        catalog = json.load(f)
    records = catalog['records']
    simulations = catalog['simulations']

    # If login is None, this creates a Login object to use
    l = login or Login()

    # Get the list of all SXS records from Zenodo
    zenodo_records = {
        r['doi_url']: r
        for r in l.search(q='communities:sxs', all_versions=True, status='published', size=9000)
    }

    # Loop through the zenodo records, ensuring that they are all present and current in the old records
    catalog_changed = False
    for r in zenodo_records:
        if r not in records or zenodo_records[r].get('modified', 0) != records[r].get('modified', 1):
            records[r] = _include_file_data(l, zenodo_records[r])
            new_simulation = _create_simulations_skeleton([records[r], ])
            simulations.update(new_simulation)
            catalog_changed = True

    if catalog_changed:
        # Loop through the dictionary we just created, and download the metadata.json for each one
        _download_sxs_metadata(l, simulations)

        # Update the modification time
        catalog['modified'] = max(catalog['modified'], max(records[doi].get('modified', '') for doi in records))

        # Write the various catalog files
        write_split_catalogs(catalog, public_dir=public_out_dir, private_dir=private_out_dir,
                             public_prefix='public_', private_prefix='private_')

        # Update the SXS-to-zenodo map
        sxs_id_to_conceptrecid = {sxsid: doi.replace('https://doi.org/10.5281/zenodo.', '')
                                  for doi in records for sxsid in [sxs_id(records[doi]['title'])] if sxsid}
        nginx_map(sxs_id_to_conceptrecid, nginx_map_file_path)

    return


def modification_time(representation_list):
    return max(r['modified'] for r in representation_list)


def sxs_metadata_file_description(representation):
    """Find metadata file from highest Lev for this simulation"""
    from os.path import basename
    files = representation.get('files', [])
    metadata_files = [f for f in files if basename(f.get('filename', '')) == 'metadata.json']
    metadata_files = sorted(metadata_files, key=lambda f: f['filename'])
    if not metadata_files:
        return None
    return metadata_files[-1]


def fetch_metadata(url, login=None, *args, **kwargs):
    """Get the json file from zenodo"""
    from .api import Login
    login = login or Login(*args, **kwargs)
    r = login.session.get(url)
    if r.status_code != 200:
        return {}
    try:
        return r.json()
    except Exception:
        return {}


def nginx_map(sxs_id_to_conceptrecid, map_file_path=None):
    """Create a mapping from SXS identifiers to Zenodo record numbers for nginx

    The input must be a catalog file.  The output is formatted for inclusion into
    an nginx configuration.  Note that this map file includes a `map_hash_max_size`
    directive, and thus must precede any `map` directives, or you will get an error
    that this directive is a duplicate (even if you never explicitly gave it
    previously).

    The output map matches the plain SXS identifier (with nothing following it).

    Parameters
    ----------
    sxs_id_to_conceptrecid : dict
        A dictionary mapping SXS IDs to Zenodo's concept-record IDs.
    map_file_path : None or string [defaults to None]
        Relative or absolute path to the nginx map file to be output.  If `None`,
        the function first attempts to write the file into
        '~/.sxs/catalog/sxs_to_zenodo.map'; if that is not possible, it tries to
        write to the same file in the current working directory.

    """
    from os.path import expanduser
    import math

    def touch(fname, times=None):
        from os import utime
        try:
            with open(fname, 'a'):
                utime(fname, times)
            return True
        except:
            pass
        return False
    # Decide on where to put the output
    if map_file_path is None:
        if touch(expanduser('~/.sxs/catalog/sxs_to_zenodo.map')):
            map_file_path = expanduser('~/.sxs/catalog/sxs_to_zenodo.map')
        elif touch('sxs_to_zenodo.map'):
            map_file_path = 'sxs_to_zenodo.map'
        else:
            raise ValueError("Cannot write to 'sxs_to_zenodo.map' file in current directory or in ~/.sxs/catalog/.")
    else:
        map_file_path = expanduser(map_file_path)
    # Figure out how to construct the output
    size = 64 * 2**math.ceil(math.log2(len(sxs_id_to_conceptrecid)+1))  # nginx needs this to know the hash size
    # Construct the output
    with open(map_file_path, 'w') as f:
        f.write("map_hash_max_size {0};\n".format(size))
        f.write("map $sxs_id $zenodo_identifier {\n")
        f.write("    default ../communities/sxs;\n")  # A bit hackish, but it gets us where we want to go
        for sxs_identifier in sorted(sxs_id_to_conceptrecid):
            f.write("    {0} {1};\n".format(sxs_identifier, sxs_id_to_conceptrecid[sxs_identifier]))
        f.write("}\n")


def resolutions_for_simulation(sxs_id, sxs_catalog):
    """List available resolutions for a given sxs_id and sxs catalog metadata

    Parameters
    ----------
    sxs_id : str
        This should be a key in the sxs_catalog['simulations'] dictionary.
    sxs_catalog : dict

    Returns
    -------
    resolutions : list
        Sorted list of integers representing the Levs available for the given SXS
        ID.

    """
    from .. import lev_number
    url = sxs_catalog.get('simulations', {}).get(sxs_id, {}).get('url', '')
    files = sxs_catalog.get('records', {}).get(url, {}).get('files', [])
    resolutions = {  # This is a set
        lev
        for f in files
        for lev in [lev_number(f.get('filename', ''))]
        if lev is not None
    }
    return sorted(resolutions)


def resolutions_for_simulations(sxs_catalog):
    """Map available SXS IDs to corresponding available resolutions

    Parameters
    ----------
    sxs_catalog : dict

    Returns
    -------
    map_sxs_id_to_resolutions : dict
        Maps each SXS ID to a sorted list of integers representing the Levs
        available for it.

    """
    return {
        sxs_id: resolutions_for_simulation(sxs_id, sxs_catalog)
        for sxs_id in sxs_catalog.get('simulations', {})
    }



# Copyright (c) 2020, Michael Boyle
# See LICENSE file for details:
# <https://github.com/sxs-collaboration/sxs/blob/main/LICENSE>

"""Interface to SXS data and utilities"""

try:
    import importlib.metadata as importlib_metadata
except ModuleNotFoundError:  # pragma: no cover
    import importlib_metadata

__version__ = importlib_metadata.version(__name__)

from . import utilities
from .utilities import (
    file_format, sxs_directory, read_config, write_config, sxs_id, lev_number,
    jit, vectorize, guvectorize
)
from .time_series import TimeSeries
from .metadata import Metadata
from .catalog import Catalog
from .horizons import Horizons, HorizonQuantities
from .waveforms import WaveformModes #, WaveformGrid, WaveformSignal
from . import catalog, metadata, horizons, waveforms
from .handlers import load, loadcontext

# The speed of light is, of course, defined to be exactly
speed_of_light = 299_792_458.0  # m/s

# By "IAU 2012 Resolution B2", the astronomical unit is defined to be exactly
astronomical_unit = 149_597_870_700.0  # m

# The parsec is, in turn, defined as "The distance at which 1 au subtends 1 arc sec: 1 au divided by
# pi/648000."  Thus, the extended-precision value of the parsec in meters is
parsec_in_meters = 3.0856775814913672789139379577965e16  # m

# The value of the solar mass parameter G*M_sun is known to higher accuracy than either of its
# factors.  The value here is taken from the publication "2018 Selected Astronomical Constants",
# which can be found at <http://asa.hmnao.com/static/files/2018/Astronomical_Constants_2018.pdf>.
# In the TDB (Barycentric Dynamical Time) time scale — which seems to be the more relevant one and
# looks like the more standard one for LIGO — we have
solar_mass_parameter = 1.32712440041e20  # m^3/s^2

# Dividing by the speed of light squared, we get the mass of the sun in meters; dividing again, we
# get the mass of the sun in seconds. For consistency with the nominal value above, we retain more
# precision than is warranted by the measurement.
m_sun_in_meters = 1476.6250385063112526099633973363  # m
m_sun_in_seconds = 4.925490949162941425997992909269e-06  # s



"""Function to inspect file formats"""


def file_format(file):
    """Return format stored in file

    If a format is specified, we assume that it is a top-level attribute or member
    named "sxs_format" or just "format".  If neither exists, return None; the
    calling function should check for this possibility.

    Parameters
    ----------
    file : file-like object, string, or pathlib.Path

    Returns
    -------
    format : {None, str}
        If the format is not specified, return None.  Otherwise, return the format
        as a string.

    """
    from pathlib import Path
    import json
    import h5py
    if isinstance(file, (str, Path)):
        path = Path(file).expanduser().resolve()
        if not path.exists():
            raise FileNotFoundError(f"Could not find {path}")
        if h5py.is_hdf5(str(path)):
            with h5py.File(str(path), "r") as f:
                return f.attrs.get("sxs_format", f.attrs.get("format", None))
        else:
            try:
                with path.open("r") as f:
                    f_json = json.load(f)
                    return f_json.get("sxs_format", f_json.get("format", None))
            except Exception as e:
                raise ValueError(f"Failed to interpret '{path}' as HDF5 or JSON file") from e
    elif hasattr(file, "read"):
        try:
            file.seek(0)
            mode = "binary" if isinstance(file.read(0), bytes) else "text"
            if mode == "binary":
                # This file is open in binary mode, so it must be an HDF5 file
                with h5py.File(file, "r") as f:
                    return f.attrs.get("sxs_format", f.attrs.get("format", None))
            else:
                # This file is open in text mode, so it must be a JSON file
                f_json = json.load(file)
                return f_json.get("sxs_format", f_json.get("format", None))
        except Exception as e:
            if mode == "binary":
                raise ValueError(
                    "\nInput `file` is open in binary mode, but is not an open HDF5 file."
                    "\nIf this is a JSON file, open it in text mode: `open(file, 'r')`"
                ) from e
            else:
                raise ValueError(
                    "\nInput `file` is open in text mode, but is not an open JSON file."
                    "\nIf this is an HDF5 file, open it in binary mode: `open(file, 'rb')`"
                ) from e
        finally:
            file.seek(0)
    else:
        raise TypeError(f"Input `file` has type {type(file)}; it should be str, pathlib.Path, or a file-like object")



class Deposit(object):

    def __init__(self, login, deposition_id=None, ignore_deletion=False):
        """Initialize a Deposit object for creating a new zenodo entry

        This object encapsulates all the actions you might want to take when creating, publishing,
        updating, or replacing an entry in zenodo.

        Parameters
        ----------
        login: Login object
            Object containing a `requests.Session` object that can successfully interact with the
            Zenodo API.  See the help string for `Login` in this module.

        deposition_id: string, int, or None [default: None]
            If present, this is used as the id of the deposit to edit.  If `None`, a new
            deposit is created.

        ignore_deletion: bool [default: False]
            If True and this object is deleted before the record is published, issue a warning
            explaining that and suggesting how it might be published by the user.

        """
        self.login = login
        self.ignore_deletion = ignore_deletion
        self._representation = {}  # Just during construction, to prevent errors if construction fails

        # Now, create or reacquire the specific deposit we're looking for
        if deposition_id is not None:
            # If the deposit id was given, check that we can access it
            url = "{0}api/deposit/depositions/{1}".format(self.base_url, deposition_id)
            r = self._get(url)
            if r.status_code != 200:
                print('The input deposition id "{0}" could not be accessed on {1}.'.format(deposition_id, url))
                try:
                    print(r.json())
                except:
                    pass
                r.raise_for_status()
                raise RuntimeError()  # Will only happen if the response was not strictly an error
        else:
            url = "{0}api/deposit/depositions".format(self.base_url)
            r = self._post(url, data="{}")
            if r.status_code != 201:
                print('Unable to create a new deposit on {0}.'.format(url))
                try:
                    print(r.json())
                except:
                    pass
                r.raise_for_status()
                raise RuntimeError()  # Will only happen if the response was not strictly an error

        # Now, using the response generated above, set some data describing this deposit
        self._representation = r.json()

    @property
    def _get(self):
        return self.login.session.get

    @property
    def _post(self):
        return self.login.session.post

    @property
    def _put(self):
        return self.login.session.put

    @property
    def _delete(self):
        return self.login.session.delete

    @property
    def refresh_information(self):
        """Retrieve current information about this Deposit from Zenodo

        This function updates this object's `representation` data, which contains information like
        the id, metadata, submission status, and various links for this deposit.  That
        information is also used by many of this object's other member functions.

        """
        url = "{0}api/deposit/depositions/{1}".format(self.base_url, self.deposition_id)
        r = self._get(url)
        if r.status_code != 200:
            print('This deposit (id "{0}") could not be accessed on {1}.'.format(self.deposition_id, url))
            try:
                print(r.json())
            except:
                pass
            r.raise_for_status()
            raise RuntimeError()  # Will only happen if the response was not strictly an error
        self._representation = r.json()
        return self.representation

    @property
    def base_url(self):
        return self.login.base_url

    @property
    def representation(self):
        return self._representation

    @property
    def metadata(self):
        return self.representation['metadata']

    @property
    def links(self):
        return self.representation['links']
    
    @property
    def deposition_id(self):
        """Return id number of this deposit"""
        return self.id

    @property
    def id(self):
        """Return id number of this deposit"""
        return self.representation['id']

    @property
    def id_latest(self):
        """Return id number of the most recent version of this deposit"""
        return self.links['latest'].split('/')[-1]

    @property
    def id_latest_draft(self):
        """Return id number of the most recent draft (unpublished) version of this deposit

        Note: There may be no draft version, in which case this function will raise a KeyError.

        """
        return self.links['latest_draft'].split('/')[-1]

    @property
    def website(self):
        """URL of web page on which this deposit is found"""
        return self.links['html']

    @property
    def website_latest(self):
        """URL of web page on which the current version of this deposit is found"""
        return self.links['latest_html']

    @property
    def website_latest_draft(self):
        """URL of web page on which the current draft (unpublished) version of this deposit is found

        Note: There may be no draft version, in which case this function will raise a KeyError.

        """
        return self.links['latest_draft_html']

    @property
    def record(self):
        """URL of API endpoint for this deposit"""
        return self.links['record']

    @property
    def record_latest(self):
        """URL of API endpoint for most recent version of this deposit"""
        return self.links['latest']

    @property
    def record_latest_draft(self):
        """URL of API endpoint for most recent draft (unpublished) version of this deposit

        Note: There may be no draft version, in which case this function will raise a KeyError.

        """
        return self.links['latest_draft']

    @property
    def state(self):
        """Current status of this deposit

        The documentation states that this may be one of:
            * inprogress: Deposit metadata can be updated. If deposit is also unsubmitted (see
              submitted) files can be updated as well.
            * done: Deposit has been published. 
            * error: Deposit is in an error state - contact Zenodo support.
        However, 'inprogress' seems to have been replaced by 'unsubmitted'.

        """
        return self.representation.get('state', 'error')
    
    @property
    def submitted(self):
        """Return True if this deposit has been submitted/published"""
        return bool(self.representation.get('submitted', False))
    
    @property
    def published(self):
        """Return True if this deposit has been submitted/published"""
        return self.submitted and self.state=='done'

    @property
    def is_latest(self):
        """Return True if this deposit is the most recent version"""
        return (self.links.get('latest', 1) == self.links.get('record', 2))

    def get_latest(self):
        """Return a new Deposit object pointing to the latest version of this deposit
        
        Note: This deposit may already be the latest version, in which case a new object pointing
        to this deposit is returned.

        """
        return self.login.deposit(self.id_latest)

    def get_latest_draft(self):
        """Return a new Deposit object pointing to the latest draft (unpublished) version of this deposit

        Note: There may be no draft version, in which case this function will raise a KeyError.

        """
        return self.login.deposit(self.id_latest_draft)

    @property
    def versions(self):
        """Get information for all versions of this deposit

        The versions are returned as a list of record data from oldest to newest.  The only change
        that this function makes to the data is to add a version number to the metadata if it is not
        present already.

        """
        url = "{0}api/deposit/depositions".format(self.base_url)
        conceptrecid = self.representation['conceptrecid']
        params = {
            'q': 'conceptrecid:{0}'.format(conceptrecid),
            'all_versions': '',
            'sort': 'version',
        }
        r = self._get(url, params=params)
        if r.status_code != 200:
            print('The list of versions for this deposit (id "{0}") could not be accessed on {1}.'.format(self.deposition_id, url))
            try:
                print(r.json())
            except:
                pass
            r.raise_for_status()
            raise RuntimeError()  # Will only happen if the response was not strictly an error
        version_list = r.json()
        for v, version in enumerate(version_list, 1):
            if 'version' not in version['metadata']:
                version['metadata']['version'] = v
        return version_list

    @property
    def title(self):
        return self.representation['title']

    @property
    def files(self):
        """Return list of dictionaries describing uploaded files

        Each file is described by a dictionary containing these keys:

            * checksum (MD5 fingerprint)
            * filename
            * filesize (in bytes)
            * id (Zenodo-generated hex id of this file)
            * links
                * download
                * self

        The last item, 'links' is another dictionary giving a couple URLs; it is not clear how to
        use these URLs.  For example, trying to get the 'download' link results in a 404 message.

        """
        return self.representation['files']

    @property
    def file_checksums(self):
        return {d['filename']: d['checksum'] for d in self.files}

    @property
    def file_sizes(self):
        return {d['filename']: d['filesize'] for d in self.files}

    @property
    def file_ids(self):
        return {d['filename']: d['id'] for d in self.files}

    @property
    def file_names(self):
        return [d['filename'] for d in self.files]

    def update_metadata(self, metadata, refresh_information=True):
        """Update this deposit with the given metadata

        The `metadata` argument should be a dictionary representing the metadata.

        Note that a deposit cannot be published unless all required metadata fields are present,
        including things like 'doi', which are automatically assigned by the system.  So if you are
        changing the metadata of a deposit, it's usually simpler to copy the metadata, modify it
        as desired, and then use this function to update it on Zenodo.


        Metadata keys and allowed values
        --------------------------------
        title: string, required

        creators: list of dictionaries, required
            Each entry should be a dictionary of these entries:
                * name: Name of creator in the format Family name, Given names
                * affiliation: Affiliation of creator (optional).
                * orcid: ORCID identifier of creator (optional).
                * gnd: GND identifier of creator (optional).

        upload_type: string, required
            * 'publication'
            * 'poster'
            * 'presentation'
            * 'dataset'
            * 'image'
            * 'video'
            * 'software'
            * 'lesson'
            * 'other'

        publication_type: string, required if 'upload_type' is 'publication'
            * 'book'
            * 'section'
            * 'conferencepaper'
            * 'article'
            * 'patent'
            * 'preprint'
            * 'deliverable'
            * 'milestone'
            * 'proposal'
            * 'report'
            * 'softwaredocumentation'
            * 'thesis'
            * 'technicalnote'
            * 'workingpaper'
            * 'other'

        image_type: string, required if if 'upload_type' is 'image'
            * 'figure'
            * 'plot'
            * 'drawing'
            * 'diagram'
            * 'photo'
            * 'other'
        
        publication_date: string [defaults to current date]
            Date of publication in ISO8601 format (YYYY-MM-DD).

        description: string, required
            Abstract or description for deposit. The following HTML tags are allowed: a, p, br,
            blockquote, strong, b, u, i, em, ul, ol, li, sub, sup, div, strike.

        access_right: string [defaults to 'open']
            * 'open'
            * 'embargoed'
            * 'restricted'
            * 'closed'

        license: string [defaults to 'cc-by' for non-datasets and 'cc-zero' for datasets]

        embargo_date: date [defaults to current date]
            When the deposited files will be made automatically made publicly available by the
            system.

        access_conditions: string, required if 'access_right' is 'restricted'
            Specify the conditions under which you grant users access to the files in your
            upload. User requesting access will be asked to justify how they fulfil the
            conditions. Based on the justification, you decide who to grant/deny access. You are not
            allowed to charge users for granting access to data hosted on Zenodo. Following HTML
            tags are allowed: a, p, br, blockquote, strong, b, u, i, em, ul, ol, li, sub, sup, div,
            strike.

        keywords: list of strings
            Free form keywords for this deposit.

        notes: string
            Additional notes. No HTML allowed.

        communities: list of dictionaries
            List of communities in which you wish the deposit to appear. The owner of the
            community will be notified, and can either accept or reject your request. Each array
            element is a dictionary with the key 'identifier' followed by the name of a Zenodo
            community.

        doi: string
            Digital Object Identifier. Did a publisher already assign a DOI to your deposited files?
            If not, leave the field empty and we will register a new DOI for you when you publish. A
            DOI allow others to easily and unambiguously cite your deposit.

        prereserve_doi: bool
            Set to `True`, to reserve a Digital Object Identifier (DOI). The DOI is automatically
            generated by our system and cannot be changed. Also, The DOI is not registered with
            DataCite until you publish your deposit, and thus cannot be used before
            then. Reserving a DOI is useful, if you need to include it in the files you upload, or
            if you need to provide a dataset DOI to your publisher but not yet publish your
            dataset. The response from the REST API will include the reserved DOI.

        related_identifiers: list of dictionaries
            Persistent identifiers of related publications and datasets. Supported identifiers
            include: DOI, Handle, ARK, PURL, ISSN, ISBN, PubMed ID, PubMed Central ID, ADS
            Bibliographic Code, arXiv, Life Science Identifiers (LSID), EAN-13, ISTC, URNs and
            URLs. Each array element is an object with the attributes:
                * identifier: The persistent identifier
                * relation: Relationship. Controlled vocabulary (isCitedBy, cites, isSupplementTo,
                  isSupplementedBy, isNewVersionOf, isPreviousVersionOf, isPartOf, hasPart,
                  compiles, isCompiledBy, isIdenticalTo, isAlternateIdentifier).

        contributors: list of dictionaries
            The contributors of the deposit (e.g., editors, data curators, etc.). Each array
            element is a dictionary with the keys:
                * name: Name of creator in the format Family name, Given names
                * type: Contributor type. Controlled vocabulary (ContactPerson, DataCollector,
                  DataCurator, DataManager, Editor, Researcher, RightsHolder, Sponsor, Other)
                * affiliation: Affiliation of creator (optional).
                * orcid: ORCID identifier of creator (optional).
                * gnd: GND identifier of creator (optional).

        references: list of strings
            List of references.

        grants: list of dictionaries
            List of OpenAIRE-supported grants, which have funded the research for this
            deposit. Each array element is an object with the key 'id' giving the grant ID.

        journal_title: string
            Journal title, if deposit is a published article.

        journal_volume: string
            Journal volume, if deposit is a published article.

        journal_issue: string
            Journal issue, if deposit is a published article.

        journal_pages: string
            Journal pages, if deposit is a published article.

        conference_title: string
            Title of conference (e.g., 20th International Conference on Computing in High Energy and
            Nuclear Physics).

        conference_acronym: string
            Acronym of conference (e.g., CHEP'13).

        conference_dates: string
            Dates of conference (e.g., 14-18 October 2013). Conference title or acronym must also be
            specified if this field is specified.

        conference_place: string
            Place of conference in the format city, country (e.g., Amsterdam, The
            Netherlands). Conference title or acronym must also be specified if this field is
            specified.

        conference_url: string
            URL of conference (e.g., http://www.chep2013.org/).

        conference_session: string
            Number of session within the conference (e.g., VI).

        conference_session_part: string
            Number of part within a session (e.g., 1).

        imprint_publisher: string
            Publisher of a book/report/chapter

        imprint_isbn: string
            ISBN of a book/report

        imprint_place: string
            Place of publication of a book/report/chapter in the format city, country.

        partof_title: string
            Title of book for chapters

        partof_pages: string
            Pages numbers of book

        thesis_supervisors: array of objects
            Supervisors of the thesis. Same format as for creators.

        thesis_university: string
            Awarding university of thesis.

        subjects: array of objects
            Specify subjects from a taxonomy or controlled vocabulary. Each term must be uniquely
            identified (e.g., a URL). For free form text, use the keywords field. Each array element
            is an object with the attributes:
                * term: Term from taxonomy or controlled vocabulary.
                * identifier: Unique identifier for term.
                * scheme: Persistent identifier scheme for id (automatically detected).

        version: string
            Version of the resource. Any string will be accepted, however the suggested format is a
            semantically versioned tag (see more details on semantic versioning at semver.org)

        language: string
            Specify the main language of the record as ISO 639-2 or 639-3 code, see Library of
            Congress ISO 639 codes list.

        """
        import json
        url = "{0}api/deposit/depositions/{1}".format(self.base_url, self.deposition_id)
        r = self._put(url, data=json.dumps({'metadata': metadata}))
        if r.status_code != 200:
            print('Updating deposit {0} failed.'.format(self.deposition_id))
            try:
                print(r.json())
            except:
                pass
            r.raise_for_status()
            raise RuntimeError()  # Will only happen if the response was not strictly an error
        if refresh_information:
            self.refresh_information
        return r

    def edit(self, refresh_information=True):
        """Unlock a previously submitted deposit for editing."""
        url = "{0}api/deposit/depositions/{1}/actions/edit".format(self.base_url, self.deposition_id)
        r = self._post(url)
        if r.status_code != 201:
            print('Unlocking deposit {0} for editing failed.'.format(self.deposition_id))
            if r.status_code == 400:
                print('Deposit state does not allow for editing (e.g., deposits in state `inprogress`).')
            if r.status_code == 409:
                print('Deposit is in the process of being integrated.  Please wait 5 minutes before trying again.')
            try:
                print(r.json())
            except:
                pass
            r.raise_for_status()
            raise RuntimeError()  # Will only happen if the response was not strictly an error
        if refresh_information:
            self.refresh_information
        return r
        
    def discard(self, refresh_information=True):
        """Discard changes in the current editing session."""
        url = "{0}api/deposit/depositions/{1}/actions/discard".format(self.base_url, self.deposition_id)
        r = self._post(url)
        if r.status_code != 201:
            print('Discarding changes from the current editing session to deposit {0} failed.'.format(self.deposition_id))
            if r.status_code == 400:
                print('Deposit is not being edited.')
            try:
                print(r.json())
            except:
                pass
            r.raise_for_status()
            raise RuntimeError()  # Will only happen if the response was not strictly an error
        if refresh_information:
            self.refresh_information
        return r

    def get_new_version(self, ignore_deletion=True, refresh_information=True):
        """Create a new Deposit object describing a new version of this deposit.

        To publish the new version, its files must differ from all previous versions.

        This action will create a new deposit, which will be a snapshot of the current resouce,
        inheriting the metadata as well as snapshot of files. The new version deposit will have a
        state similar to a new, unpublished deposit, most importantly its files will be modifiable
        as for a new deposit.

        Only one unpublished new version deposit can be available at any moment, i.e.: calling new
        version action multiple times will have no effect, as long as the resulting new version
        deposit from the first call is not published or deleted.

        """
        self.register_new_version(refresh_information=refresh_information)
        return self.login.deposit(self.id_latest_draft, ignore_deletion=ignore_deletion)
        
    def register_new_version(self, refresh_information=True):
        """Create a new version of a deposit.

        To publish the new version, its files must differ from all previous versions.

        This action will create a new deposit, which will be a snapshot of the current resouce,
        inheriting the metadata as well as snapshot of files. The new version deposit will have a
        state similar to a new, unpublished deposit, most importantly its files will be modifiable
        as for a new deposit.

        Only one unpublished new version deposit can be available at any moment, i.e.: calling new
        version action multiple times will have no effect, as long as the resulting new version
        deposit from the first call is not published or deleted.

        NOTE: The response body of this action is NOT the new version deposit, but the original
        resource. The new version deposit can be accessed through the "latest_draft" under
        "links" in the response body.

        """
        url = "{0}api/deposit/depositions/{1}/actions/newversion".format(self.base_url, self.deposition_id)
        r = self._post(url)
        if r.status_code != 201:
            print('Failed to register new version of deposit {0}.'.format(self.deposition_id))
            if r.status_code == 403:
                print('This deposit may not have been published, in which case new versions are not allowed.')
            try:
                print(r.json())
            except:
                pass
            r.raise_for_status()
            raise RuntimeError()  # Will only happen if the response was not strictly an error
        if refresh_information:
            self.refresh_information
        return r

    def delete_file(self, file_name, refresh_information=True):
        matching_files = [(i, f) for i, f in enumerate(self.files) if f['filename'] == file_name]
        if not matching_files:
            print('File name "{0}" not found on Zenodo.'.format(file_name))
            raise ValueError(file_name)
        file_index, file = matching_files[0]
        file_name = file['filename']
        file_id = file['id']
        url = '{0}api/deposit/depositions/{1}/files/{2}'.format(self.base_url, self.id, file_id)
        r = self._delete(url)
        if r.status_code != 204:
            print('Deleting file "{0}" from deposit "{1}" failed.'.format(file_name, self.id))
            if r.status_code == 403:
                print('Server replied with "Forbidden: Deleting an already published deposition file."')
                print('Try get_new_version, and then delete the file from that version.')
            if r.status_code == 404:
                print('File id "{0}" does not exist in deposition "{1}".'.format(file_id, self.id))
                print('Try refreshing the local information in this Deposit object, and trying again as relevant.')
            try:
                print(r.json())
            except:
                pass
            r.raise_for_status()
            raise RuntimeError()  # Will only happen if the response was not strictly an error
        if refresh_information:
            self.refresh_information
        else:
            self.representation['files'].pop(file_index)
        return r

    def upload_file(self, path, name=None, relpath_start=None, skip_checksum=False, refresh_information=True):
        """Upload a single file to the deposit

        The current list of files uploaded to Zenodo is checked.  If `name` is in that list, the MD5
        checksum of `path` is evaluated and compared to the MD5 checksum of the file on Zenodo.  If
        they match, the upload is skipped and `None` is returned.

        Parameters
        ----------
        path: string
            Relative or absolute path to the file to upload
        name: string or None [default: None]
            Name of the file as it should appear in the deposit.  This can be the same as or
            different from the path, and can contain directories.  If this is None, the name will be
            the relative path from `relpath_start` to the file itself.  Note that if the absolute
            path to `relpath_start` is not contained within the absolute path to `path`, then that
            name would start with '../' or something.  All such parts are removed from the `name`.
        relpath_start: string [default: current directory]
            Relative or absolute path at which to start the relative path to the `name` if required.
        skip_checksum: bool [default: False]
            If True, ignore the checksum and always upload the file.  If False, see if any file with
            exactly this name and checksum has been uploaded to Zenodo; if so, don't bother
            uploading it again.

        """
        import os
        from ...utilities import md5checksum
        if relpath_start is None:
            relpath_start = os.curdir
        if name is None:
            abspath = os.path.abspath(path)
            absstart = os.path.abspath(relpath_start)
            name = os.path.relpath(abspath, absstart)
            pardir_sep = os.pardir + os.sep
            while name.startswith(pardir_sep):
                name = name[len(pardir_sep):]
        if not skip_checksum:
            file_checksums = self.file_checksums
            if name in file_checksums:
                if md5checksum(path) == file_checksums[name]:
                    print('File {0} has already been uploaded.  Skipping this upload.'.format(name))
                    return None
        if name in self.file_ids:
            self.delete_file(name, refresh_information=False)
        url = '{0}/{1}'.format(self.links['bucket'], name)
        r = self._put(url, data=open(path, 'rb'),  headers={"Content-Type": "application/octet-stream"})
        if r.status_code != 200:
            print('Uploading {0} to deposit {1} failed.'.format(path, self.deposition_id))
            print('Upload url was {0}.'.format(url))
            if r.status_code == 400:
                if os.stat(path).st_size == 0:
                    print('This file has size zero, which leads to an error response.')
            try:
                print(r.json())
            except:
                pass
            r.raise_for_status()
            raise RuntimeError()  # Will only happen if the response was not strictly an error
        if refresh_information:
            self.refresh_information
        return r

    def rename_file(self, old_file_name, new_file_name, refresh_information=True):
        import json
        file_ids = self.file_ids
        if old_file_name not in file_ids:
            print('File name "{0}" not found on Zenodo.'.format(old_file_name))
            raise ValueError(old_file_name)
        file_id = file_ids[old_file_name]
        url = '{0}api/deposit/depositions/{1}/files/{2}'.format(self.base_url, self.id, file_id)
        rename_data = {'filename': new_file_name}
        r = self._put(url, data=json.dumps(rename_data))
        if r.status_code != 200:
            print('Renaming file "{0}" to "{1}" in deposit "{2}" failed.'.format(old_file_name, new_file_name, self.id))
            if r.status_code == 403:
                print('Server replied with "Forbidden: Renaming an already published deposition file."')
                print('Try get_new_version, and then delete the file from that version.')
            if r.status_code == 404:
                print('File id "{0}" does not exist in deposition "{1}".'.format(file_id, self.id))
                print('Try refreshing the local information in this Deposit object, and trying again as relevant.')
            try:
                print(r.json())
            except:
                pass
            r.raise_for_status()
            raise RuntimeError()  # Will only happen if the response was not strictly an error
        if refresh_information:
            self.refresh_information
        return r
        
    def upload_all_files(self, top_directory, exclude=[], skip_checksum=False, refresh_information=True):
        """Recursively upload all files found in `top_directory`

        This function simply calls the `sxs.utilities.find_files` function, and then calls
        `upload_file` for each one.  See those functions for explanations of the parameters to this
        function.

        """
        from ...utilities import find_files
        paths_and_names = find_files(top_directory, exclude=exclude)
        for path, name in paths_and_names:
            print("Uploading\n    {0}\nas\n    {1}".format(path, name))
            self.upload_file(path, name=name, skip_checksum=skip_checksum, refresh_information=False)
            print("Upload succeeded\n")
        if refresh_information:
            self.refresh_information

    def prereserve_doi(self, refresh_information=True):
        """Pre-reserve the DOI for an unpublished deposit, so that it does not change upon publication

        Note that this is undocumented behavior of the zenodo API, so it may change without notice;
        this function was reverse-engineered by analyzing requests made by the web interface.  The
        API itself is not explicit about how this could work, but the approach that is vaguely
        suggested does not work.

        """
        from copy import deepcopy
        metadata = deepcopy(self.metadata)
        metadata.update({'doi': metadata['prereserve_doi']['doi']})
        return self.update_metadata(metadata, refresh_information=refresh_information)
    
    def publish(self, refresh_information=True):
        """Publish this deposit on Zenodo.

        In order to be published successfully, a deposit must contain at least one file, and all
        of the required metadata fields must be present.  In particular if you have changed the
        metadata, be sure that you have not removed one of the fields that is automatically
        assigned, like 'doi' or 'prereserve_doi', etc.

        Note that you will not be able to change the files after publishing, unless you create a new
        version of this deposit, which will result in a new DOI -- though anyone looking for this
        deposit will see a notice that there is a newer version.  You will still be able to edit
        the metadata (including the description) without changing the DOI.

        """
        url = '{0}api/deposit/depositions/{1}/actions/publish'.format(self.base_url, self.deposition_id)
        r = self._post(url)
        if r.status_code != 202:
            print('Publishing deposit {0} failed.'.format(self.deposition_id))
            if r.status_code == 500:
                print('Server returned the code "500 Internal Server Error".')
                print('This can have any number of causes, but frequently it is because not')
                print('all of the required information is present.  In particular, every')
                print('deposit must contain at least one file, and all of its metadata')
                print('must be present.  See the warning in `Deposit.update_metadata`.')
            try:
                print(r.json())
            except:
                pass
            r.raise_for_status()
            raise RuntimeError()  # Will only happen if the response was not strictly an error
        if refresh_information:
            self.refresh_information
        return r
    
    def delete_deposit(self, confirmed=False):
        """Permanently delete this deposit from Zenodo

        If you are sure that you want to delete this deposit, you can pass `confirmed=True`.
        Otherwise, you will be prompted to input 'yes' in order to complete the deletion.  This
        prompt will only last for 60 seconds.  If no input is given, we assume that the answer is
        'no', a warning is printed, and the program continues.

        """
        if not confirmed:
            import sys, select
            timeout = 60
            print("Please confirm that you want to delete the deposit {0}.".format(self.deposition_id))
            print("You have {0} seconds to confirm by entering 'yes'.".format(timeout))
            i, o, e = select.select([sys.stdin], [], [], timeout)
            if not i or sys.stdin.readline().strip().lower() != 'yes':
                print('No confirmation received.  Aborting deletion of zenodo deposit {0}.'.format(self.deposition_id))
                raise RuntimeError('No confirmation')
        url = '{0}api/deposit/depositions/{1}'.format(self.base_url, self.deposition_id)
        r = self._delete(url)
        if r.status_code != 204:
            print('Deleting deposit {0} failed.'.format(self.deposition_id))
            try:
                print(r.json())
            except:
                pass
            r.raise_for_status()
            raise RuntimeError()  # Will only happen if the response was not strictly an error
        return r

    def __del__(self):
        if not self.published and not self.ignore_deletion:
            from textwrap import dedent
            from warnings import warn

            warning = r"""
            The Zenodo Deposit object has not been published.  The deposit id is '{deposition_id}'.
            If you want to publish this deposit, you can do it manually from the website by
            going to

                {base_url}deposit/{deposition_id}

            filling in any remaining information that is needed, and clicking "Publish".  If you do not
            want to publish it at any point, you can go to that page and click "Delete".

            You can also perform these actions using this interface by running

                >>> from sxs.zenodo import Login
                >>> l = Login(<YourLoginInfo>)
                >>> d = l.deposit({deposition_id})
                >>> # d.upload_file(...), d.update_metadata(...), etc.
                >>> d.publish()

            or

                >>> d.delete()

            """.format(deposition_id=self.deposition_id, base_url=self.base_url)
            warn(dedent(warning))
